@@ -34,12 +34,12 @@
 #include <errno.h>
 
 #include "cfg.h"
-#include "cfg_global.h"
 #include "cfg_fmt.h"
 #include "cfg_main.h"
 #include "cfg_data.h"
 #include "cfg_node.h"
 #include "cfg_syn.h"
+#include "cfg_global.h"
 #define _CFG_VERS_C_AS_HEADER_
 #include "cfg_vers.c"
 #undef  _CFG_VERS_C_AS_HEADER_
@@ -73,11 +73,18 @@ cfg_rc_t cfg_create(cfg_t **cfg)
     return CFG_OK;
 }
 
+static cfg_rc_t cfg_destroy_node(cfg_t *cfg, cfg_node_t *node, void *ctx)
+{
+    cfg_node_destroy(cfg, node);
+    return CFG_OK;
+}
+
 cfg_rc_t cfg_destroy(cfg_t *cfg)
 {
     if (cfg == NULL)
         return CFG_ERR_ARG;
-    /* FIXME: on each node, free token string first */
+    if (cfg->root != NULL)
+        cfg_node_apply(cfg, cfg->root, NULL, NULL, cfg_destroy_node, NULL);
     cfg_grid_destroy(cfg->grid_nodes);
     cfg_grid_destroy(cfg->grid_tokens);
     free(cfg);
