@@ -32,12 +32,12 @@
 #include <ctype.h>
 
 #include "cfg.h"
-#include "cfg_global.h"
 #include "cfg_grid.h"
 #include "cfg_node.h"
 #include "cfg_fmt.h"
 #include "cfg_syn.h"
 #include "cfg_buf.h"
+#include "cfg_global.h"
 
 /* prototypes for Flex-generated scanner */
 extern int  cfg_syn_lex_init(void *);
@@ -57,6 +57,8 @@ cfg_rc_t cfg_syn_import(
     size_t err_len)
 {
     cfg_syn_ctx_t ctx;
+    cfg_buf_t *buf;
+    cfg_rc_t rc;
     void *yyscan;
 
     /* argument sanity checking */
@@ -67,6 +69,10 @@ cfg_rc_t cfg_syn_import(
     cfg_syn_lex_init(&yyscan);
     cfg_syn_set_extra(&ctx, yyscan);
 
+    /* initialize temporary buffer context */
+    if ((rc = cfg_buf_create(&buf)) != CFG_OK)
+        return rc;
+
     /* establish our own context which is passed
        through the parser and scanner */
     ctx.inputptr = in_ptr;
@@ -75,6 +81,7 @@ cfg_rc_t cfg_syn_import(
     ctx.cfg      = cfg;
     ctx.node     = NULL;
     ctx.rv       = CFG_OK;
+    ctx.buf      = buf;
     ctx.err_buf  = err_buf;
     ctx.err_len  = err_len;
     ctx.yyscan   = yyscan;
@@ -83,6 +90,9 @@ cfg_rc_t cfg_syn_import(
     if (cfg_syn_parse(&ctx))
         ctx.rv = (ctx.rv == CFG_OK ? CFG_ERR_INT : ctx.rv);
 
+    /* destroy temporary buffer */
+    cfg_buf_destroy(buf);
+
     /* destroy scanner */
     cfg_syn_lex_destroy(yyscan);
 
@@ -92,6 +102,37 @@ cfg_rc_t cfg_syn_import(
     return ctx.rv;
 }
 
+/* helper function for copying out input with escaping of characters */
+static size_t cfg_syn_error_cpyout(char *out_ptr, const char *in_ptr, size_t in_len)
+{
+    size_t out_len;
+    char c;
+
+    out_len = 0;
+    while (in_len-- > 0) {
+        c = *in_ptr++;
+        switch (c) {
+            case '\n': *out_ptr++ = '\\'; *out_ptr++ = 'n'; out_len += 2; break;
+            case '\r': *out_ptr++ = '\\'; *out_ptr++ = 'r'; out_len += 2; break;
+            case '\t': *out_ptr++ = '\\'; *out_ptr++ = 't'; out_len += 2; break;
+            case '\b': *out_ptr++ = '\\'; *out_ptr++ = 'b'; out_len += 2; break;
+            case '\f': *out_ptr++ = '\\'; *out_ptr++ = 'f'; out_len += 2; break;
+            default: {
+                if (!isprint((int)c)) {
+                    *out_ptr++ = '\\';
+                    *out_ptr++ = '?';
+                    out_len += 2;
+                }
+                else {
+                    *out_ptr++ = c;
+                    out_len++;
+                }
+            }
+        }
+    }
+    return out_len;
+}
+
 /* remember a parsing error (used internally) */
 void cfg_syn_error(cfg_syn_ctx_t *ctx, cfg_rc_t rv, YYLTYPE *loc, const char *fmt, ...)
 {
@@ -142,29 +183,26 @@ void cfg_syn_error(cfg_syn_ctx_t *ctx, cfg_rc_t rv, YYLTYPE *loc, const char *fm
     }
 
     /* extract token context with mark token borders */
-    if ((cpBuf = malloc((size_t)((cpE-cpP)+2+1))) == NULL)
+    if ((cpBuf = malloc((size_t)(((cpE-cpP)*2)+2+1))) == NULL)
         return;
     cp = cpBuf;
-    n = cpF-cpP;
-    memcpy(cp, cpP, n); cp += n;
+    cp += cfg_syn_error_cpyout(cp, cpP, cpF-cpP);
     *cp++ = '<';
-    n = cpL-cpF;
-    memcpy(cp, cpF, n); cp += n;
+    cp += cfg_syn_error_cpyout(cp, cpF, cpL-cpF);
     *cp++ = '>';
-    n = cpE-cpL;
-    memcpy(cp, cpL, n); cp += n;
+    cp += cfg_syn_error_cpyout(cp, cpL, cpE-cpL);
     *cp++ = '\0';
 
     /* remember parsing error: part I (context part) */
     cfg_fmt_sprintf(ctx->err_buf, ctx->err_len,
-                    "line %d, column %d: `%s'",
+                    "line %d, column %d: `%s': ",
                     line, column, cpBuf);
     free(cpBuf);
 
     /* remember parsing error: part II (parsing part) */
     n = strlen(ctx->err_buf);
     va_start(ap, fmt);
-    cfg_fmt_vsprintf(ctx->err_buf+n, sizeof(ctx->err_len)-n, fmt, ap);
+    cfg_fmt_vsprintf(ctx->err_buf+n, ctx->err_len-n, fmt, ap);
     va_end(ap);
 
     return;
